# Copyright (C) 2017  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import json

import botogram

from .buffer import InvalidBufferTokenError


CLEANUP_AFTER = 30 * 60


def inline_keyboard(buttons):
    """Create an inline keyboard"""
    buttons_tg = []

    for row in buttons:
        inner = []
        for text, callback in row:
            inner.append({"text": text, "callback_data": callback})
        buttons_tg.append(inner)

    result = lambda: None
    result.serialize = lambda: {"inline_keyboard": buttons_tg}

    return result


class PostComponent(botogram.Component):
    component_name = "post"

    def __init__(self, db, buffer, auth):
        self.db = db
        self.buffer = buffer
        self.auth = auth

        self.add_timer(2 * 60, self.cleanup)
        self.add_timer(5 * 60, self.post_to_telegram)
        self.add_command("post", self.post_command)

    def cleanup(self):
        """Clean up old records"""
        max = time.time() - CLEANUP_AFTER
        self.db.update("DELETE FROM post_pending WHERE created_at < ?", max)

    def post_to_telegram(self, bot):
        """Post queued messages to Telegram"""
        # Return if there isn't a valid Buffer token
        try:
            self.buffer.access_token()
        except InvalidBufferTokenError:
            return

        channel = self.db.query(
            "SELECT value FROM kw WHERE key = 'telegram_channel';"
        )

        # If there isn't a channel linked discard all the pending posts
        if not channel:
            self.db.update("DELETE FROM to_telegram;")
            return
        channel = channel[0][0]

        to_telegram = [d[0] for d in self.db.query(
            "SELECT post_id FROM to_telegram;"
        )]
        for post_id in to_telegram:
            data = self.buffer.request("get", "updates/%s.json" % post_id)

            # For example if the update was deleted
            if "error" not in data:
                if data["status"] != "sent":
                    continue

                message = data["text"]
                try:
                    link = data["media"]["expanded_link"]
                except KeyError:
                    pass
                else:
                    if link:
                        message += "\n\n%s" % link

                bot.chat(channel).send(message)

            # This will delete the update if the message was sent to the
            # channel or there was an error with it
            self.db.update(
                "DELETE FROM to_telegram WHERE post_id = ?;", post_id
            )

    def post_command(self, chat, message):
        """Invia un nuovo post"""
        try:
            self.buffer.access_token()
        except InvalidBufferTokenError:
            message.reply("‼️ *Errore:* token di Buffer non valido!")
            return

        # Check if this is a reply
        if message.reply_to_message is None:
            message.reply("‼️ *Rispondi* al messaggio contenente il post!")
            return

        message_id = message.reply_to_message.message_id

        # Check if there is a publishing in progress for this post
        already_exists = self.db.query(
            "SELECT message FROM post_pending WHERE message = ?",
            message_id
        )
        if already_exists:
            message.reply_to_message.reply(
                "‼️ *Errore:* stai già per pubblicare questo post."
            )
            return

        profiles = self.buffer.request("get", "profiles.json")
        for profile in profiles:
            self.db.update(
                "INSERT INTO post_pending (message, social, social_id, "
                "social_pretty, social_name, post_there, created_at) "
                "VALUES (?, ?, ?, ?, ?, ?, ?);",
                message_id, profile["service"], profile["_id"],
                profile["formatted_service"], profile["formatted_username"],
                False, int(time.time()),
            )

        channel = self.db.query(
            "SELECT value FROM kw WHERE key = 'telegram_channel';"
        )
        if channel:
            self.db.update(
                "INSERT INTO post_pending (message, social, social_id, "
                "social_pretty, social_name, post_there, created_at) "
                "VALUES (?, ?, ?, ?, ?, ?, ?);",
                message_id, "telegram", "__telegram__", "Telegram",
                channel[0][0], False, int(time.time()),
            )

        message.reply_to_message.reply(
            "Seleziona i social dove vuoi pubblicare il post:",
            extra=self.get_keyboard(message_id),
        )

    def process_callbacks(self, bot, chains, update):
        """Process incoming callbacks"""
        query = update.callback_query
        action, data = query.data.split(":", 1)

        # Authenticate the request
        self.auth.check(query.message.chat)

        if action == "now":
            if query.sender not in query.message.chat.admins:
                bot.api.call("answerCallbackQuery", {
                    "callback_query_id": query.id,
                    "text": "Devi essere admin per poter eseguire questa "
                            "azione.",
                    "show_alert": True,
                })
                return

        if action == "toggle":
            message, social = data.split(":")
            message = int(message)

            data = self.db.query(
                "SELECT post_there FROM post_pending WHERE message = ? AND "
                "social = ?;",
                message, social,
            )
            if not data:
                query.message.edit("‼️ *Operazione scaduta!*")
                return

            current = data[0][0]
            self.db.update(
                "UPDATE post_pending SET post_there = ? WHERE message = ? AND "
                "social = ?;",
                not current, message, social
            )

            kb = self.get_keyboard(message)
            bot.api.call("editMessageReplyMarkup", {
                "chat_id": query.message.chat.id,
                "message_id": query.message.message_id,
                "reply_markup": json.dumps(kb.serialize()),
            })

        elif action in ("queue", "top", "now"):
            message = int(data)

            to_socials = [d[0] for d in self.db.query(
                "SELECT social_id FROM post_pending WHERE message = ? AND "
                "post_there = 1;", message
            )]

            telegram = False
            if "__telegram__" in to_socials:
                telegram = True
                to_socials.remove("__telegram__")

            if not to_socials:
                bot.api.call("answerCallbackQuery", {
                    "callback_query_id": query.id,
                    "text": "Devi selezionare almeno un social che non sia "
                            "Telegram!",
                    "show_alert": True
                })
                return

            data = {
                "text": query.message.reply_to_message.text,
                "profile_ids[]": to_socials,
            }
            if action == "now":
                data["now"] = True
            elif action == "top":
                data["top"] = True

            res = self.buffer.request("post", "updates/create.json", data=data)
            if res["success"]:
                # This will be sent to the user
                socials = [
                    "<code>- </code>%s: <b>%s</b>" % (d[0], d[1])
                    for d in self.db.query(
                        "SELECT social_pretty, social_name FROM post_pending "
                        "WHERE message = ? AND post_there = 1;", message
                )]

                query.message.edit("\n".join([
                        "✅ <b>Post inviato con successo!</b>"
                    if action == "now" else
                        "✅ <b>Post messo in coda!</b>",
                    "",
                    "Account social selezionati:",
                ] + socials), syntax="html")

                # Queue the update also for Telegram
                if telegram:
                    channel = self.db.query(
                        "SELECT value FROM kw WHERE key = 'telegram_channel';"
                    )[0][0]

                    # If the post should be published now, publish it instantly
                    if action == "now":
                        message = res["updates"][0]["text"]
                        try:
                            link = res["updates"][0]["media"]["expanded_link"]
                        except KeyError:
                            pass
                        else:
                            if link:
                                message += "\n\n%s" % link

                        bot.chat(channel).send(message)
                    else:
                        post_id = res["updates"][0]["id"]
                        self.db.update(
                            "INSERT INTO to_telegram (post_id) VALUES (?);",
                            post_id
                        )
            else:
                query.message.edit("\n".join((
                    "‼️ *Impossibile inviare il post:*",
                    res["message"],
                )))

            self.db.update(
                "DELETE FROM post_pending WHERE message = ?;", message
            )

        bot.api.call("answerCallbackQuery", {
            "callback_query_id": query.id,
        })

    def get_keyboard(self, message_id, edit=False):
        """Prepare the content of the /post command"""
        pending = self.db.query(
            "SELECT social, social_pretty, social_name, post_there FROM "
            "post_pending WHERE message = ? ORDER BY social_pretty;",
            message_id,
        )

        buttons = []
        for one in pending:
            buttons.append([(
                "%s%s: %s" % (
                    "✅ " if one[3] else "",
                    one[1],
                    one[2],
                ),
                "toggle:%s:%s" % (message_id, one[0])
            )])

        buttons += [[
            ("Metti in coda", "queue:%s" % message_id),
            ("Pubblica per primo", "top:%s" % message_id),
        ], [
            ("Pubblica ora", "now:%s" % message_id),
        ]]

        return inline_keyboard(buttons)
